//  acordsub.cpp 

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

/* Includes */
#ifdef WIN32
#include <afxmt.h>
#include "largeint.h"
#include <wchar.h>
#include <windows.h>
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCpi.h>
#include <BipCni.h>
#include "BipSampPluginUtil.h"

#include "parsubs.h"
#include "trace.h"
#include "mtdstruc.h"
#include "metamem.h"
#include "metadata.h"
#include "miscsubs.h"
#include "acordsub.h"
#include "common.h"

// Parser exception codes (built by message compiler)
#include "acorderr.h"

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#ifdef WIN32
#define ACORDMHGL			L"1MHG"
#define ACORDMTGL			L"3MTG"
#define ACORDTRGL			L"2TRG"
#define ACORDTCGL			L"2TCG"
#define ACORDTRANNAMEL		L"Tran"
#define ACORDGROUPVERNAME	L"Header.Version"
#define ACORD_2TCGL			L"_2TCG"
#define ACORDSTD			L"STVNO"
#define ACORDVERNO			L"VERNO"
#endif

#define MSGHDR				"MessageHeader"
#define ACORDTRANNAME		"Tran"
#define ACORDMHGSTD			"910"
#define ACORDMHGOFS			109
#define ACORDVERSIONSEG		"2TCG"
#define ACORDMHG			"1MHG"
#define ACORDMTG			"3MTG"
#define ACORDTRG			"2TRG"
#define ACORDTCG			"2TCG"
#define ACORDTRANNAME		"Tran"
#define ACORDTRGOFS			10
#define ACORDSTDNOSIZE		3
#define ACORDVERNOSIZE		2
#define ACORDVERSIONOFFSET	10
#define ACORDTCGSTDLEN		100
#define ACORDMINLENGTH		10
#define GROUPNAMELEN		4
#define GROUPLENGTHLEN		3
#define GROUPVERSIONOFS		8

#ifdef _DEBUG
static char Level[]="acordsub.cpp V1.0 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="acordsub.cpp V1.0 Release version ("__DATE__" "__TIME__")";
#endif

	extern FILETBL* fileTable;		// File table to cache recent files 

#ifdef _DEBUG
	extern int debug;
	extern int traceParser;
	extern int traceWrite;
	extern int traceModule;
	extern int traceMgmt;
#endif

	extern int     displayTruncated;	// Display fields truncated

	int     badMessageCount=0;		// Number of messages processed    
	int     segmentCount=0;			// Number of segments processed

int	getSegmentLength(const CciChar * buffer, const int offset)

{
	int	seglen;
	int rc;
	char charname[GROUPNAMELEN + 1];
	char charofs[10];
	char charlen[4];
	char errmsg[512];

	/* Initialize work area */
	memset(charlen, 0, sizeof(charlen));

	/* convert the 3 byte length from Unicode to ASCII */
	cciUcsToMbs(&rc, buffer + GROUPNAMELEN, charlen, GROUPLENGTHLEN, DEF_CODEPAGE);

	/* Convert it to an integer */
	seglen = atoi(charlen);

	/* Check if it is long enough */
	if (seglen < ACORDMINLENGTH)
	{
		// Increase the invalid message counter
		badMessageCount++;

		/* Length is too short - throw an exception */
		cciUcsToMbs(&rc, buffer, charname, GROUPNAMELEN, DEF_CODEPAGE);
		charname[GROUPNAMELEN] = 0;
		charofs[0] = 0;
		itoa(offset, charofs, 10);
		sprintf(errmsg, "Invalid segment length %s for segment %s", charlen, charname);

		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "getSegmentLength",
						  ERRDICT,
						  ACORD_BAD_LENGTH,
						  errmsg,
						  charlen,
						  charname,
						  charofs,
						  0);
	}

	return seglen;
}

/**************************************************************/
/*                                                            */
/* Routine to convert the 2 digit length in a compressed      */
/* message sequence into a integer value.                     */
/*                                                            */
/**************************************************************/

int	convCompLen(const char * charlen, int codepage)

{
	int	length=0;
	int	i;
	int rc;
	char ch;
	CciChar Ucharlen[3];
	char	Acharlen[3];

	// Get the character length in the default code page,
	// in case the input is not in ASCII
	cciMbsToUcs(&rc, charlen, Ucharlen, sizeof(Ucharlen) - 1, codepage);
	Ucharlen[sizeof(Ucharlen) - 1] = 0;
	cciUcsToMbs(&rc, Ucharlen, Acharlen, sizeof(Acharlen) - 1, DEF_CODEPAGE);

	for (i=0; i<2; i++)
	{
		// Multiply by 16
		length = length << 4;

		// handle the hex digit
		// do we have a letter or a number?
		ch = Acharlen[i];
		if ((ch <= '9') && (ch >= '0'))
		{
			// number
			length += ch & 15;
		}
		else
		{
			// letter
			// make sure it is upper case
			ch = ch & 0x0DF;
			length += ch - 'A' + 10;
		}
	}

	return length;
}

/**************************************************************/
/*                                                            */
/* Routine to extra the compressed character and the number   */
/* of compressed characters from a 4 byte compression         */
/* sequence.  The character and length will be returned in    */
/* the provided areas.  The input data is assumed to be in    */
/* the provided codepage.                                     */
/*                                                            */
/**************************************************************/

void getValues(const char * inBuf, CciChar * compChar, int * compLen, int codepage)

{
	int rc;
	CciChar UtempSeq[4];
	char	tempSeq[4];
	char	tempBuf[4];

	// clear workareas
	memset(UtempSeq, 0, sizeof(UtempSeq));
	memset(tempSeq, 0, sizeof(tempSeq));
	memset(tempBuf, 0, sizeof(tempBuf));
	memcpy(tempBuf, inBuf + 1, 3);

	// Translate the next three characters to Unicode and 
	// back to the native code page
	cciMbsToUcs(&rc, tempBuf, UtempSeq, sizeof(UtempSeq), codepage);

	if (rc == CCI_SUCCESS) 
	{
		cciUcsToMbs(&rc, UtempSeq, tempSeq, sizeof(tempSeq), DEF_CODEPAGE);
	}

	if (rc == CCI_SUCCESS)
	{
		(*compLen) = convCompLen(inBuf + 1, codepage);
		compChar[0] = UtempSeq[2];
	}
	else
	{
		(*compLen) = 0;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to decompress a message segment that has been      */
/* compressed using the ACORD character compression           */
/* algorithm.  Any sequences of characters beginning with a   */
/* x'FA' character will be expanded.  The other areas of the  */
/* message will be copied.                                    */
/*                                                            */
/* The input characters will also be translated to Unicode    */
/* when they are copied to the output buffer.                 */
/*                                                            */
/**************************************************************/

void deCompress(const char * inBuf, 
				CciChar * outBuf, 
				const int inLen, 
				const int codepage)

{
	int rc;
	int	i=0;
	int	j=0;
	int	complen;
	int count;
	const char * ptr;
	CciChar * outPtr;
	CciChar compChar;

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: -> deCompress(): inLen=%d\n", inLen);
	}
#endif

	// initialize the output pointer
	outPtr = outBuf;

	while (i < inLen)
	{
		// initialize character counter and pointer
		count = 0;
		ptr = inBuf + i;

		// Scan for a compression sequence
		while ((i < inLen) && ((const unsigned char) inBuf[i] != HEX_FA))
		{
			count++;
			i++;
		}

		// did we find any uncompressed characters?
		if (count > 0)
		{
			// move the uncompressed characters to the output buffer,
			// translating them to Unicode
			cciMbsToUcs(&rc, ptr, outPtr, count, codepage);
			outPtr += count;
		}

		// Now, check if we found a compression character
		if ((i < inLen) && ((const unsigned char) inBuf[i] == HEX_FA))
		{
			// get the compression length and character
			getValues(inBuf + i, &compChar, &complen, codepage);

			// Fill in the compressed character
			j = 0;
			while (complen > 0)
			{
				outPtr++[0] = compChar;
				complen--;
			}

			// skip over the compression sequence
			i += 4;
		}
	}

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: <- deCompress():\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to check if a message buffer contains a compressed */
/* message.  If character compression has been performed on   */
/* the message in question, then it will contain one or more  */
/* x'FA' characters and must be decompressed.                 */
/*                                                            */
/**************************************************************/

int checkCompress(char * buffer, const int length, const int codepage)

{
	int				extra=0;
	int				complen;
	unsigned char *	ptr;
	unsigned char *	endptr;

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: -> checkCompress(): length=%d\n", length);
	}
#endif

	// get pointers to the beginning and end of the buffer
	ptr = (unsigned char *) buffer;
	endptr = ptr + length;

	// Search for a compression character
	while (ptr < endptr)
	{
		if (ptr[0] != HEX_FA)
		{
			// translate any binary zeros to blanks
			if (ptr[0] == 0)
			{
				ptr[0] = ' ';
			}

			// go on to the next character
			ptr++;
		}
		else
		{
			complen = convCompLen((char *) ptr + 1, codepage);
			extra += complen - 4;
			ptr += 4;
		}
	}

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: <- checkCompress(): extra=%d\n", extra);
	}
#endif

	return extra;
}

/**************************************************************/
/*                                                            */
/* Routine to skip any new line or carriage returns between   */
/* message segments.                                          */
/*                                                            */
/**************************************************************/

void skipCRLF(PARSER_CONTEXT_ST* context, int* length, CciChar** buffer)

{
	// skip any new line or carriage return characters found between segments
	while (((*length) < context->iLength) && 
			((*buffer[0]) < chSpace))
	{
		(*buffer)++;
		(*length)++;
	} // while
}

/**************************************************************/
/*                                                            */
/* Routine to validate if this is a valid ACORD AL3 message.  */
/*                                                            */
/* To be considered a valid message, the first segment must   */
/* be a message header group, the second segment must be a    */
/* transaction header group and the last group must be a      */
/* message trailer group.  All of the length fields in the    */
/* group headers must be valid and the total length of the    */
/* length fields must match the length of the message.  If    */
/* the message is obviously invalid, then it a parser         */
/* exception will be thrown.                                  */
/*                                                            */
/**************************************************************/

void checkMessageFormat(PARSER_CONTEXT_ST* context)

{
	CciChar* buffer;
	int length=0;
	int	seglen;
	int rc;
	char charlen1[32];
	char charlen2[32];
	char segname[5];
	char errmsg[512];
	CciChar seg[GROUPNAMELEN + 1];

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: -> checkMessageFormat()\n");
	}
#endif

	// clear the work area
	memset(seg, 0, sizeof(seg));

	// get the buffer address
	buffer = context->iBuffer;

	// Check for a message header at the beginning of the message
	memcpy(seg, buffer, GROUPNAMELEN * sizeof(CciChar));
	if (CciCharCmp(seg, ACORDMHGL) != 0)
	{
		// Increase the invalid message counter
		badMessageCount++;

		// prepare an error message
		strcpy(errmsg, "No message header group found");

#ifdef _DEBUG
		if (debug > 0) 
		{
			Log("%s\n", errmsg);
		}
#endif

		// We don't have a message header group
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  ACORD_NO_MHG,
						  errmsg,
						  0);
	}

	/* Go on to the next segment */
	seglen = getSegmentLength(buffer, length);
	buffer += seglen;
	length += seglen;

#ifdef _DEBUG
		if (traceParser > 0) 
		{
			Log(" checking segment %S at offset 0 length %d total length 0\n", 
				seg, seglen);
		}
#endif

	// skip any new line or carriage return characters found between segments
	// the original buffer must be checked
	skipCRLF(context, &length, &buffer);

	// Check for a transaction 
	memcpy(seg, buffer, GROUPNAMELEN * sizeof(CciChar));
	if (CciCharCmp(seg, ACORDTRGL) != 0)
	{
		// Increase the invalid message counter
		badMessageCount++;

		// prepare an error message
		strcpy(errmsg, "No transaction header group found");

#ifdef _DEBUG
		if (debug > 0) 
		{
			Log("%s\n", errmsg);
		}
#endif

		// We don't have a transaction header group
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  ACORD_NO_TRG,
						  errmsg,
						  0);
	}

	// Find the message trailer group, checking the segment lengths
	while ((length < context->iLength) && (CciCharCmp(seg, ACORDMTGL) != 0))
	{
			/* Go on to the next segment */
		seglen = getSegmentLength(buffer, length);

		cciUcsToMbs(&rc, buffer, segname, GROUPNAMELEN, DEF_CODEPAGE);
		segname[GROUPNAMELEN] = 0;

#ifdef _DEBUG
		if (traceParser > 0) 
		{
			Log(" checking segment %s at offset %d length %d total length %d\n", 
				segname, buffer - context->iBuffer, seglen, length);
		}
#endif

		if ((seglen > (context->iLength - length)) || (seglen < ACORDMINLENGTH))
		{
			// Invalid length found in group header
			sprintf(errmsg, "Invalid length (%d) found at offset (%d)", 
					seglen, buffer - context->iBuffer);

#ifdef _DEBUG
			if (debug > 0) 
			{
				Log("%s\n", errmsg);
			}
#endif

			// get the length and offset as character fields
			itoa(length, charlen1, 10);
			itoa(buffer - context->iBuffer, charlen2, 10);

			cciThrowException(&rc, 
							  CCI_PARSER_EXCEPTION, 
							  __FILE__, 
							  __LINE__, 
							  "checkMessageFormat",
							  ERRDICT,
							  ACORD_BAD_LENGTH,
							  errmsg,
							  charlen1,
							  segname,
							  charlen2,
							  0);
		}

		// go on to the next group
		buffer += seglen;
		length += seglen;

		// skip any new line or carriage return characters found between segments
		skipCRLF(context, &length, &buffer);

		// capture the next name of the next group
		memcpy(seg, buffer, GROUPNAMELEN * sizeof(CciChar));
	}

	// Check for a message trailer group 
	if (CciCharCmp(seg, ACORDMTGL) != 0)
	{
		// Increase the invalid message counter
		badMessageCount++;

		// prepare an error message
		strcpy(errmsg, "No message trailer group found");

#ifdef _DEBUG
		if (debug > 0) 
		{
			Log("%s\n", errmsg);
		}
#endif

		// We don't have a message trailer group
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  ACORD_NO_MTG,
						  errmsg,
						  0);
	}

	// Get the total length
	seglen = getSegmentLength(buffer, length);

#ifdef _DEBUG
		if (traceParser > 0) 
		{
			Log(" checking segment %s at offset %d length %d total length %d\n", 
				segname, buffer - context->iBuffer, seglen, length);
		}
#endif

	buffer += seglen;
	length += seglen;

	// skip any new line or carriage return characters at the end of the message
	skipCRLF(context, &length, &buffer);

	// Check if the total length is correct
	if (length != context->iLength)
	{
		// Total length of all groups does not match message size
		sprintf(errmsg, "Lengths do not match %d %d", length, context->iSize);

#ifdef _DEBUG
		if (debug > 0) 
		{
			Log("%s\n", errmsg);
		}
#endif

		// Get the lengths as character strings
		charlen1[0] = 0;
		itoa(length, charlen1, 10);
		charlen2[0] = 0;
		itoa(context->iLength, charlen2, 10);

		// Increase the invalid message counter
		badMessageCount++;

		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "checkMessageFormat",
						  ERRDICT,
						  ACORD_LENGTH_ERR,
						  errmsg,
						  charlen1,
						  charlen2,
						  0);
	}

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: <- checkMessageFormat()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to find the group version number in an output      */
/*  segment.  The group version number is a child of the      */
/*  header, which is the first child of the segment.          */
/*  The element name is assumed to be "Version".              */
/*  If the element is not found, then a value of zero is      */
/*  returned.  The version number must be a digit from 1 to   */
/*  9 or a capital letter from A to Z.                        */
/*                                                            */
/*  The first child of the root must be the "Header" element. */
/*                                                            */
/**************************************************************/

char findGroupVersion(CciElement* segmentRoot)

{
	int			rc;
	char		groupVer=0;
	char		value[3];
	CciElement*	verElement;

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: -> findGroupVersion() segmentRoot=0x%x\n", segmentRoot);
	}
#endif

	verElement = findElementByName(ACORDGROUPVERNAME, segmentRoot);
	if (verElement != CCI_NULL_ADDR)
	{
		// we found the element, now get the value
		rc = findCharacterValue(value, sizeof(value), verElement, DEF_CODEPAGE);
		if (rc == CCI_SUCCESS)
		{
			// make sure the version is valid
			if (((value[0] >= '1') && (value[0] <= '9')) || ((value[0] >= 'A') && (value[0] <= 'Z')))
			{
				// Yes, it appears valid - pass it back
				groupVer = value[0];
			}
		}
	}
	else
	{
#ifdef _DEBUG
		if (traceWrite > 0) 
		{
			Log(" Header.Version element not found\n");
		}
#endif
	}

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: <- findGroupVersion() groupVer(%c)\n", groupVer);
	}
#endif

	return groupVer;
}

/**************************************************************/
/*                                                            */
/* Routine to find the version number in the output parse     */
/*  tree 1MHG element.                                        */
/*                                                            */
/*  The version number for the message header and trailer     */
/*   is carried in the 1MHG element itself.  The standard     */
/*   number is not present but is assumed to be 910.  This    */
/*   number is used if no group version number is present.    */
/*                                                            */
/*  This routine is also used for the 2TRG element, which     */
/*   also carries a standard version number within the        */
/*   group.                                                   */
/*                                                            */
/**************************************************************/

void buildHeaderMetaLevel(char * headerPtr, CciElement* MHG_Element)

{
	CciElement*		verno;
	int				rc;
	char			verchar[3];

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: -> buildHeaderMetaLevel() 1MHG_Element=0x%x\n", MHG_Element);
	}
#endif

	// look for the VERNO element
	verno = findElementByName(ACORDVERNO, MHG_Element);
	if (verno != CCI_NULL_ADDR)
	{
		// Element found, now get the value
		rc = findCharacterValue(verchar, sizeof(verchar), verno, DEF_CODEPAGE);
		if (rc == CCI_SUCCESS)
		{
			// copy in the standard number (910 for message header and trailer)
			memcpy(headerPtr, ACORDMHGSTD, sizeof(ACORDMHGSTD) - 1);

			// copy in the version number for this standard
			setNumericValue(headerPtr + ACORDSTDNOSIZE, 
							ACORDVERNOSIZE, 
							verchar, 
							strlen(verchar));

			// terminate the string
			headerPtr[ACORDSTDNOSIZE + ACORDVERNOSIZE] = 0;
		}
		else
		{
#ifdef _DEBUG
			if (traceModule > 0) 
			{
				Log("Unable to get character value\n");
			}
#endif
		}
	}
	else
	{
		// Element not found
#ifdef _DEBUG
		if (traceModule > 0) 
		{
			Log("VERNO element not found\n");
		}
#endif
	}

#ifdef _DEBUG
	if (traceModule > 0) 
	{
		Log("PLUGIN: <- buildHeaderMetaLevel() headerPtr(%s)\n", headerPtr);
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to initialize an output buffer segment, setting    */
/*  all variable values to defaults.                          */
/*                                                            */
/* This routine requires two data areas, one for the data in  */
/* unicode and one for the data in character format.  The     */
/* size of the work areas is initially set to 256 characters  */
/* but the areas will be increased if an individual item is   */
/* larger.  The larger area will then be kept for the         */
/* duration of the segment initialization.                    */
/*                                                            */
/**************************************************************/

void initialize(char * outBuf, int fileIdx, int firstVar, int lastVar)

{
	int	i;
	int j;
	int k;
	int	occur;
	int	length;
	int	var1;
	int	var2;
	int addlOfs;
	char *	ptr;
	char value=0;
	char * bufOfs;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> initialize() outBuf=0x%x fileIdx=%d firstVar=%d lastVar=%d\n", outBuf, fileIdx, firstVar, lastVar);
	}
#endif

	// check if we have an initialization string to use directly
	if (fileTable[fileIdx].initString > 0)
	{
		// copy the initialization string to the output buffer
		ptr = fileTable[fileIdx].initString;
		memcpy(outBuf, ptr, fileTable[fileIdx].dataLength);

#ifdef _DEBUG
		if (traceWrite) 
		{
			Log("Buffer initialized with initialization string - length %d\n", 
				fileTable[fileIdx].dataLength);
		}
#endif
	}
	else
	{
		i=firstVar;
		while(i < lastVar)
		{
			// calculate the number of occurences of this variable
			// this will return a 1 if there is no occurs count
			occur = 1;
			if (fileTable[fileIdx].fvar[i].Occur > 1)
			{
				occur = fileTable[fileIdx].fvar[i].Occur;
			}

			if (fileTable[fileIdx].fvar[i].OccurMax > occur)
			{
				occur = fileTable[fileIdx].fvar[i].OccurMax;
			}

			// check if we have an elementary variable
			if (fileTable[fileIdx].fvar[i].FirstChild == 0)
			{
				// initialize this variable,
				// handling any occurs clauses
				bufOfs = outBuf + fileTable[fileIdx].fvar[i].Offset;

				for (j=0; j < occur; j++);
				{
					// Get the data into a work area
					length = fileTable[fileIdx].fvar[i].Length;
				
#ifdef _DEBUG
					if (traceWrite) 
					{
						Log("PLUGIN: -> Setting value for variable (%s) occur=%d of %d\n", 
							fileTable[fileIdx].namePtr + fileTable[fileIdx].fvar[i].Nameptr, j, occur);
					}
#endif

					// check for numeric rather than character data
					if ((fileTable[fileIdx].fvar[i].Flags & FLAGS_NUMERIC) > 0)
					{
						setNumericValue(bufOfs, length, &value, 0);
					}
					else
					{
						setDisplayValue(bufOfs,
										length,
										&value,
										0);
					}

					bufOfs += length;
				}
			}
			else
			{
				// must be higher level group item
				// only need to initialize if subject of occurs
				if (occur > 1)
				{
#ifdef _DEBUG
					if (traceWrite) 
					{
						Log("PLUGIN: -> Handling variable (%s) occur=%d\n", 
							fileTable[fileIdx].namePtr + fileTable[fileIdx].fvar[i].Nameptr, occur);
					}
#endif

					// initialize this variable,
					// handling any occurs clauses
					addlOfs = 0;
					var1 = fileTable[fileIdx].fvar[i].FirstChild - 1;
					var2 = fileTable[fileIdx].fvar[i].LastChild;
					length = fileTable[fileIdx].fvar[i].Length;

					k = 0;
					while (k < occur)
					{
						// We will recursively enter the initialize routine
						initialize(outBuf + addlOfs, fileIdx, var1, var2); 

						addlOfs += length;
						k++;
					}

					// We already processed the entire group - no need to do it again
					i = var2 - 1;
				}
			}

			i++;
		}
	}	

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- initialize() outBuf=0x%x fileIdx=0%d\n", outBuf, fileIdx);
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to update the value of a particular cobol variable */
/*  from data in the logical message.                         */
/*                                                            */
/**************************************************************/

void updateValue(char * dataloc, 
				 VAR_FILE_REC* fvar, 
				 int varnum, 
				 CciElement* element,
				 int codepage)

{
	int			RC;
	int			maxLength;
	CciValueType datatype;
	CciSize		size;
	CciElementType type;
	CciElement*	valueElement;
	CciInt		tempInt;
	char		decimal[64];
	CciChar		value[512];
	CciChar*	bigBuffer;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> updateValue() element=0x%x, varnum(%d)\n", element, varnum);
	}
#endif

	type = cpiElementType(&RC, element);
	if (RC == CCI_SUCCESS)
	{
		if (type == CCI_ELEMENT_TYPE_NAME)
		{
			valueElement = cpiFirstChild(&RC, element);
		}
		else
		{
			if ((type == CCI_ELEMENT_TYPE_NAME_VALUE) || (type == CCI_ELEMENT_TYPE_VALUE))
			{
				valueElement = element;
			}
			else
			{
#ifdef _DEBUG
				// log the error
				if (traceWrite) 
				{
					Log("Unrecognized element type 0x%x\n", type);
				}
#endif
				// Unknown element type - give up
				return;
			}
		}
	}
	else
	{
#ifdef _DEBUG
		// log the error
		if (traceWrite) 
		{
			Log("Unable to get element type RC=%d\n", RC);
		}
#endif

		return;
	}

	// Get the data type
	datatype = cniElementValueType(&RC, valueElement);
	if (RC != CCI_SUCCESS)
	{
#ifdef _DEBUG
		// log the error
		if (traceWrite) 
		{
			Log("Unable to get element value type, RC = %d\n", RC);
		}
#endif

		return;
	}

	if (datatype == CCI_VALUE_TYPE_CHAR)
	{
		// Character data,
		// find out how large the data is
		size = cpiElementCharacterValue(&RC, valueElement, value, sizeof(value));
	}
	else
	{
		// We can convert an integer to character data and use that
		if (datatype == CCI_VALUE_TYPE_INTEGER)
		{
			// Get the data as an integer
			tempInt = cpiElementIntegerValue(&RC, valueElement);
			_i64toa(tempInt, decimal, 10);
			setNumericValue(dataloc, fvar[varnum].Length, decimal, strlen(decimal));

#ifdef _DEBUG
		if (traceWrite) 
		{
			Log("  integer value %s length %d\n", decimal, fvar[varnum].Length);
		}
#endif
			// All done, return
			return;
		}
		else
		{
			// Can't handle this data type
#ifdef _DEBUG
			if (debug) 
			{
				Log("Unable to convert value to character for data type %d\n", datatype);
			}
#endif
			return;
		}
	}

	if (RC == 0)
	{
		// move the data into the output area
		// calculate the maximum length allowed
		maxLength = size - 1;

		// data will be truncated
		if (maxLength > fvar[varnum].Length)
		{
			// capture as much data as we can and
			// add a trace record
			maxLength = fvar[varnum].Length;
			displayTruncated++;
#ifdef _DEBUG
			if (traceWrite) 
			{
				Log("data truncated - maxLength = %d, size = %d\n", maxLength,size);
			}
#endif
		}
		else
		{
#ifdef _DEBUG
			if (traceWrite) 
			{
				Log("   var length %d size %d value %S\n", 
					fvar[varnum].Length, size, value);
			}
#endif
		}

		// Move the data into the output buffer
		size = cciUcsToMbs(&RC, value, (char *)dataloc, maxLength, codepage);
	}
	else
	{
		// Was our buffer big enough to hold the value?
		if (RC == CCI_BUFFER_TOO_SMALL)
		{
			// no, try to get a larger buffer
			bigBuffer = (CciChar *)malloc(size * sizeof(CciChar));

			// check if the malloc worked
			if (bigBuffer == NULL)
			{
				// no, malloc failed for bigger buffer
#ifdef _DEBUG
				if (debug) 
				{
					Log("malloc failed for large buffer - size %d\n", size);
				}
#endif
				// give up
				return;
			}
			else
			{
				// malloc worked
				// try again with the larger buffer
				size = cpiElementCharacterValue(&RC, valueElement, bigBuffer, size);
				if (RC == 0)
				{
					// calculate the maximum length and 
					// move the data to the output buffer
					maxLength = size;
					if (size > maxLength)
					{
						// data will be truncated
						// capture as much data as we can and
						// add a trace record
						maxLength = fvar[varnum].Length;

						// Keep a count of how many fields were truncated
						displayTruncated++;
#ifdef _DEBUG
						if (traceWrite) 
						{
							Log("data truncated - maxLength = %d, size = %d\n", maxLength,size);
						}
#endif
					}

					// Move the data into the output buffer
					size = cciUcsToMbs(&RC, value, (char *)dataloc, maxLength, codepage);
				}
				else
				{
					// still didn't work, write a trace record and move on
#ifdef _DEBUG
					if (debug) 
					{
						Log("unable to get character value, RC = %d, size = %d\n", RC, size);
					}
#endif
				}

				// release the storage area
				free(bigBuffer);
			}
		}
		else
		{
#ifdef _DEBUG
			if (debug) 
			{
				Log("Unable to get value for element(ox%x) RC=%d\n", valueElement, RC);
			}
#endif
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- updateValue() element=0x%x\n", element);
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to process all the children of a particular name   */
/*  element.                                                  */
/*                                                            */
/**************************************************************/

CciElement * processLevel(int firstchild,
						  VAR_FILE_REC* fvar,
						  CciChar* UnameArea,
						  CciElement* element,
						  int codepage,
						  char * outBuf)

{
	int	varnum;
	int	occur;
	int	occOffset;
	int	i;
	int rc;
	CciElement* child;
	CciElement* saveElement;
	CciChar* Uptr;
	CciSize size;
	CciChar Uname[512];

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1); 
		Log("PLUGIN: -> processLevel() variable=(%d) element=%S (0x%x)\n", 
			firstchild, Uname, element);
	}
#endif

	varnum = firstchild;
	while (varnum != 0)
	{
#ifdef _DEBUG
		if (traceWrite) 
		{
			size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1); 
			Log(" processing variable(%d) name(%S) element %S (0x%x)\n", varnum, 
				UnameArea + fvar[varnum].Unameptr, Uname, element);
		}
#endif

		// remember the current element location in the logical message
		// in case we do not find this value
		saveElement = element;

		Uptr = UnameArea + fvar[varnum].Unameptr;
		element = findNextElement(element, Uptr);

		if (element != CCI_NULL_ADDR)		
		{
			// found the element we are looking for
			// get the occurs value
			// make sure we don't this element again
			saveElement = CCI_NULL_ADDR;

			// get the number of occurences of this variable
			occur = fvar[varnum].Occur;

			i = 0;
			occOffset = 0;
			// Find out if this is a group or elementary item
			if (fvar[varnum].FirstChild == 0)
			{
				while ((i < occur) && (element != CCI_NULL_ADDR))
				{
#ifdef _DEBUG
					if (traceWrite) 
					{
						size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1); 
						Log(" updating variable(%S) with element %S (0x%x) i=%d of %d\n", 
							UnameArea + fvar[varnum].Unameptr, Uname, element, i, occur);
					}
#endif
					// elementary item
					updateValue(outBuf + fvar[varnum].Offset + occOffset,
								fvar,
								varnum,
								element,
								codepage);
					// next variable in occurs
					occOffset += fvar[varnum].Length;
					i++;
#ifdef _DEBUG
					if (traceWrite) 
					{
						size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1); 
						Log(" element %S (0x%x) occOffset %d i=%d of %d\n", 
							Uname, element, occOffset, i, occur);
					}
#endif

					// we used the current element
					// move on to the next element
					element = cpiNextSibling(&rc, element);

					// remember this element
					saveElement = element;

					// does this element occur more than once and
					// if so, is this still the correct element?
					if ((element != CCI_NULL_ADDR) && (i < occur))
					{
						// skip any elements that don't match
						element = findNextElement(element, Uptr);

						// get the name of the element
						Uname[0] = 0;
						size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1);
						if (CciCharCmp(Uname, Uptr) != 0)
						{
							// couldn't find any more elements, so try the 
							// next entry in the metadata, starting with
							// the first entry we haven't used
							element = saveElement;
							i = occur;
						}
					}
				}
			}
			else
			{
				while (i < occur)
				{
					// group item - check for any child elements
					child = cpiFirstChild(&rc, element);

					if ((rc == 0) && (child != CCI_NULL_ADDR))
					{
						child = processLevel((fvar[varnum].FirstChild),
											 fvar,
											 UnameArea,
											 child,
											 codepage,
											 outBuf + occOffset);

#ifdef _DEBUG
						if (traceWrite) 
						{
							size = cpiElementName(&rc, child, Uname, sizeof(Uname) - 1); 
							Log(" child %S (0x%x) occOffset %d i=%d of %d\n", 
								Uname, child, occOffset, i, occur);
						}
#endif
					}
					else
					{
						// No children at this level
						// enter a trace record
#ifdef _DEBUG
						if (traceWrite) 
						{
							Log(" no children for element(0x%x)\n", element);
						}
#endif
					}

					// go on to the next entry
					// next variable in occurs
					occOffset += fvar[varnum].Length;

					// go on to the next occurs item
					i++;

					// we used this element
					element = cpiNextSibling(&rc, element);

					// remember this element
					saveElement = element;

					// does this element occur more than once and
					// if so, is this still the correct element?
					if ((element != CCI_NULL_ADDR) && (i < occur))
					{
						// skip any elements that don't match
						element = findNextElement(element, Uptr);

						// get the name of the element
						Uname[0] = 0;
						size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1);
						if (CciCharCmp(Uname, Uptr) != 0)
						{
							// couldn't find any more elements, so try the 
							// next entry in the metadata, starting with
							// the first entry we haven't used
							element = saveElement;
							i = occur;
						}
					}
				} // while
			}
		}
		else
		{
			// didn't find the element, go back to the original element,
			// since this element may just be missing
			element = saveElement;

#ifdef _DEBUG
			if (traceModule) 
			{
				size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1); 
				Log("*** didn't find element for %S return to element %s (0x%x) \n", 
					Uptr, Uname, element);
			}
#endif
		}

		// go on to the next variable at this level
		varnum = (fvar[varnum].NextSibling);
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, element, Uname, sizeof(Uname) - 1); 
		Log("PLUGIN: <- processLevel(), element %S (0x%x)\n", Uname, element);
	}
#endif

	return element;
}

/**************************************************************/
/*                                                            */
/* Routine to create the output buffer for a given output     */
/*  segment.                                                  */
/*                                                            */
/* If a variable is not found, then the data field will be    */
/*  set to a default value.                                   */
/*                                                            */
/* The metadata file will be assumed to be the definition     */
/* of the output.  Only variables that can be matched to a    */
/* field in the metadata will be used.  Furthermore, the      */
/* logical message data must be in the same order as the copy */
/* book data, or fields in the logical message will be        */
/* skipped. If there are multiple occurences, they must be    */
/* contiguous.  If there are more occurences in the logical   */
/* message data, the extra occurences will be ignored.        */
/* Fields which are in the logical message but not in the     */
/* copy book will be ignored.                                 */
/*                                                            */
/**************************************************************/

int createOutputSegment(CciParser* parser,
						PARSER_CONTEXT_ST* context,
						char* buffer,
						const int bufsize,
						const char* segName,
						const char * versionPtr) 

{
	int		buflen;
	int		rc;
	int		fileIdx=0;			// Index to file table
	char *	outBuf;				// pointer to output buffer
	CciChar * UoutBuf;			// pointer to Unicode work area
	char	charofs[32];		// Work area to build offset value
	char	errmsg[512];
	CciChar	Ubuffer[1000];		// Unicode temporary buffer
	char	groupVer;			// Group version number, if present
	CciElement* nextChild;		// working variable
#ifdef _DEBUG
	CciChar	name[256];
	CciSize	size;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(name, 0, sizeof(name));
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1);
		Log("PLUGIN: -> createOutputSegment() %s context->eParent %S (0x%x) version(%s)\n", 
			segName, name, context->eParent, versionPtr);
	}
#endif

	// see if there is a group version number with this segment
	groupVer = findGroupVersion(context->eParent);

#ifdef _DEBUG
	if (traceWrite) 
	{
		Log(" Processing segment(%s) groupVer(%c)\n",segName, groupVer);
	}
#endif

	// get the meta data for this segment type, passing in the 
	// segment identifier
#ifdef _DEBUG
	fileIdx = getMetaData(traceWrite, versionPtr, segName, groupVer);
#else
	fileIdx = getMetaData(0, versionPtr, segName, groupVer);
#endif

	if (fileIdx < 0)
	{
		// didn't find the file meta data
		// write an error message
		sprintf(errmsg, "Error 110 getting meta data for segment %s\n", segName);

#ifdef _DEBUG
		if (debug)
		{
			Log(errmsg);
		}
#endif

		charofs[0] = 0;
		itoa(context->iOffset, charofs, 10);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "createOutputSegment",
						  ERRDICT,
						  ACORD_META_NOT_FOUND,
						  errmsg,
						  segName,
						  charofs,
						  0);

		// We should never get here but just in case
		return 0;
	}
	else
	{
		// found the meta data for this segment
		// Now, we can get the output data area and
		// initialize it
		buflen = fileTable[fileIdx].fvar[0].Length;
		if (buflen > bufsize)
		{
			outBuf = (char *)malloc(buflen);
			UoutBuf = (CciChar *) malloc(buflen);
		}
		else
		{
			outBuf = buffer;
			UoutBuf = Ubuffer;
		}

		// Now initialize the fields to default values
		// This is important, since some field values may
		// not be in the logical message model data
		initialize(outBuf, fileIdx, 0, fileTable[fileIdx].maxvar);

		// check if the segment name and length have been 
		// filled in by the initialization string
		if (outBuf[0] <= ' ')
		{
			// Initialize the first seven bytes to a valid 
			// group identifier and length
			memcpy(outBuf, segName, GROUPNAMELEN);

			// convert the length to a character format
			itoa(buflen, charofs, 10);

			// initialize the length field
			memset(outBuf + GROUPNAMELEN, '0', GROUPLENGTHLEN);
			memcpy(outBuf + GROUPNAMELEN + GROUPLENGTHLEN - strlen(charofs), charofs, strlen(charofs));

			// set the group version if there is one
			if (groupVer > ' ')
			{
				outBuf[GROUPVERSIONOFS] = groupVer;
			}
		}

		// Now, we have to check the code page of the output
		if (context->iCodePage != DEF_CODEPAGE)
		{
			// The output is supposed to be a different code page
			// We must translate the initialized buffer to Unicode
			// and back to the proper code page
			cciMbsToUcs(&rc, outBuf, UoutBuf, buflen, DEF_CODEPAGE);
			cciUcsToMbs(&rc, UoutBuf, outBuf, buflen, context->iCodePage);
		}

		// Update the output data area with any values
		// in the logical message model tree
		// The elements will be used to navigate through
		// the tree.  Value elements will be assumed to be
		// the only child of a name element.  Only name
		// elements can have children.
		nextChild = cpiFirstChild(&rc,context->eParent);

		if (nextChild != CCI_NULL_ADDR)
		{
			processLevel(1,
						 fileTable[fileIdx].fvar, 
						 fileTable[fileIdx].UnamePtr, 
						 nextChild,
						 context->iCodePage,
						 outBuf);
		}

		// Decrement the file usage counter
#ifdef _DEBUG
		decrementUseCount(traceWrite, fileIdx);
#else
		decrementUseCount(0, fileIdx);
#endif

		// append this data to the output buffer
		cpiAppendToBuffer(&rc, parser, (char *) outBuf, buflen);

		// Keep track of total output length
		context->iOffset += buflen;

		if (rc == 0)
		{
			// successfully appended
#ifdef _DEBUG
			if (traceWrite)
			{
				Log("Data appended to buffer - length %d at offset %d\n", 
					buflen, context->iOffset - buflen);
			}
#endif
		}
		else
		{
			// failed
#ifdef _DEBUG
			if (debug)
			{
				Log("Append failed - rc %d length %d at offset %d\n", 
					rc, buflen, context->iOffset);
			}
#endif
		}

		// release the storage we used to create this segment, if necessary
		if (buflen > bufsize)
		{
			free(outBuf);
			free(UoutBuf);
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		memset(name, 0, sizeof(name));
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1);
		Log("PLUGIN: <- createOutputSegment() length %d context->eParent %S (0x%x)\n", 
			buflen, name, context->eParent);
	}
#endif

	return buflen;
}
					
/**************************************************************/
/*                                                            */
/* Routine to create the output buffer from the logical       */
/*  message.                                                  */
/*                                                            */
/* The segment type must be checked for a 2TCG segment.  If   */
/* this is a 2TCG segment, then the version data must be      */
/* captured and used for the other segments within this       */
/* particular transaction.                                    */
/*                                                            */
/**************************************************************/

int createOutputMessage(CciParser* parser, PARSER_CONTEXT_ST* context)

{
	int			rc;
	int			i;
	int			total=0;
	CciSize		size;
	char*		vPtr;
	char		segName[GROUPNAMELEN + 1];
	char		buffer[1000];		// output data buffer, to capture 2TCG data
	char		longName[256];
	CciChar		UtempName[256];
	CciChar		UtempStd[ACORDTCGSTDLEN + 1];

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> createOutputMessage() parser=0x%x\n", parser);
	}
#endif

	// Get the root element of the body
	context->eBody = cpiRootElement(&rc, parser);
	context->eParent = CCI_NULL_ADDR;

	// Get the first child of the body, if we found the body
	if ((context->eBody != CCI_NULL_ADDR) && (rc == 0))
	{
		context->eParent = cpiFirstChild(&rc, context->eBody);
	}

	// Process all the main sections of the body
	while ((context->eParent != CCI_NULL_ADDR) && (rc == 0))
	{
		// point to the version area by default
		vPtr = context->msgVersion;

		// get the name of this element
		cpiElementName(&rc, context->eParent, UtempName, sizeof(UtempName) - 1);

		// did we get the name?
		if (rc == CCI_SUCCESS)
		{
			// convert from unicode to ASCII
			size = cciUcsToMbs(&rc, UtempName, longName, sizeof(longName) - 1, DEF_CODEPAGE);
		}

		if (rc != CCI_SUCCESS)
		{
			// couldn't find a name, so make sure the name is NULL
			longName[0] = 0;
		}

		// There are two possibilities for the name
		// There is a set mapping of longer names to the short
		// ACORD AL3 group names.  The other supported naming
		// convention is to use the ACORD group name preceded by
		// an underscore character.
		// check if the name starts with an underscore
		if (longName[0] == '_')
		{
			// get rid of the leading underscore
			for (i=0; i < GROUPNAMELEN; i++)
			{
				segName[i] = longName[i+1];

				// truncate to 4 characters
				segName[GROUPNAMELEN] = 0;
			}
		}
		else
		{
			getGroupName(segName, longName);
		}

#ifdef _DEBUG
	if (traceWrite) 
	{
		Log("  long name is %s, group id %s\n", longName, segName);
	}
#endif

		if (rc == 0)
		{
			if (strcmp(segName, ACORDTCG) == 0)
			{
				vPtr = context->trnVersion;
			}

			// Check for a message header group
			if (strcmp(segName, ACORDMHG) == 0)
			{
				// we must find the version level for this element, if present
				buildHeaderMetaLevel(context->hdrVersion, context->eParent);
				vPtr = context->hdrVersion;
			}

			// Check for a message trailer group
			if (strcmp(segName, ACORDMTG) == 0)
			{
				// Just use whatever we found in the message header group
				vPtr = context->hdrVersion;
			}

			// Check for a transaction header group
			if (strcmp(segName, ACORDTRG) == 0)
			{
				// we must find the version level for this element, if present
				buildHeaderMetaLevel(context->trnVersion, context->eParent);
				vPtr = context->trnVersion;
			}
		}

		// check if this is a transaction element
		if (strcmp(longName, ACORDTRANNAME) == 0)
		{
			// This is the start of a transaction
#ifdef _DEBUG
			if (traceWrite) 
			{
				Log("Processing transaction\n");
			}
#endif
			// move down to the next level
			context->eParent = cpiFirstChild(&rc, context->eParent);

			// rememeber we are in a transaction
			context->iInTran = 1;
		}
		else
		{
			// not a transaction element - create an output segment
			total += createOutputSegment(parser, context, buffer, sizeof(buffer), segName, vPtr);

			if (strcmp(segName, ACORDTCG) == 0)
			{
				// save the standard version and level information,
				// making sure to translate it to ASCII
				memset(UtempStd, 0, sizeof(UtempStd));
				memset(context->msgVersion, 0, sizeof(context->msgVersion));
				cciMbsToUcs(&rc, buffer + ACORDVERSIONOFFSET, UtempStd, ACORDTCGSTDLEN, context->iCodePage);
				cciUcsToMbs(&rc, UtempStd, context->msgVersion, ACORDTCGSTDLEN, DEF_CODEPAGE);
#ifdef _DEBUG
				if (traceWrite) 
				{
					Log(" msgVersion saved (%S)\n", UtempStd);
				}
#endif
			}

			// check if there are more elements
			if ((cpiNextSibling(&rc, context->eParent) == CCI_NULL_ADDR) &&
				(context->iInTran == 1))
			{
				// processed all the elements in this transaction
				// we need to move on to the next transaction or the trailer element
				context->eParent = cpiParent(&rc, context->eParent);
				context->iInTran = 0;
			}

			// process the next element at this level
			context->eParent = cpiNextSibling(&rc, context->eParent);
		}

		// check if we are done with this transaction
		if ((context->eParent == CCI_NULL_ADDR) && (context->iInTran == 1))
		{
			// processed all the elements at this level
			// we need to move on to the next transaction or the trailer element
			context->eParent = cpiParent(&rc, context->eParent);
			context->iInTran = 0;
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- createOutputMessage() length %d\n", total);
	}
#endif

	// return total bytes written
	return total;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Sets level complete and makes the parent   */
/* the current element.                                                       */
/*                                                                            */
/******************************************************************************/

void finishHeaderElement(PARSER_CONTEXT_ST* context)

{
	int                rc;
#ifdef _DEBUG
	CciSize size;
	CciChar name[512];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1); 
		Log("PLUGIN: -> finishHeaderElement(); eParent=%S (0x%x) bits(%d %d)\n", 
			name, context->eParent, completePrev, completeNext);
	}
#endif

	// set the last child complete bit for the current element
	cpiSetElementCompleteNext(&rc, context->eParent, COMPLETE_BIT);

	// restore the main element as the current parent
	context->eParent = cpiParent(&rc, context->eParent);

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1); 
		Log("PLUGIN: <- finishHeaderElement(); eParent=%S (0x%x) bits(%d %d)\n",
			name, context->eParent, completePrev, completeNext);
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Creates Specified name tag.                */
/*                                                                            */
/******************************************************************************/

CciElement * createHeaderElement(CciParser* parser,
								 PARSER_CONTEXT_ST* context,
								 const CciChar* tagName)

{
	CciElement*        element;
	int                rc;
#ifdef _DEBUG
	CciSize size;
	CciChar name[512];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1); 
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		Log("PLUGIN: -> createHeaderElement(); tagName=%S parent=%S (0x%x)\n", 
			tagName, name, context->eParent, completePrev, completeNext);
	}
#endif

	/* Create a new name element for the tag */
	element = cpiCreateAndInitializeElement(&rc, 
					parser,
					CCI_ELEMENT_TYPE_NAME,
					tagName,
					COMPLETE_BIT,
					0);

	/* Add the element as the last child of the current parent */
	cpiAddAsLastChild(&rc, context->eParent, element);

	// mark the first child pointer of the parent as complete
	cpiSetElementCompletePrevious(&rc, context->eParent, COMPLETE_BIT);

	// make the new element the current parent
	context->eParent=cpiLastChild(&rc, context->eParent);

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, element);
		completeNext = cpiElementCompleteNext(&rc, element);
		Log("PLUGIN: <- createHeaderElement(); created element=%S (0x%x) bits (%d %d)\n", 
			tagName, element, completePrev, completeNext);
	}
#endif

	return element;
}

/**************************************************************/
/*                                                            */
/* Routine to add an elementary item to the current logical   */
/*  message.                                                  */
/*                                                            */
/* All values will be entered as character values.  Any       */
/* necessary conversion from the original data type will be   */
/* performed.                                                 */
/*                                                            */
/**************************************************************/

void addItem(CciParser* parser,
			 PARSER_CONTEXT_ST* context,
			 const CciChar* tagName,
			 const CciChar *dataIn,
			 const int length,
			 const int usage)

{
	int	rc;
	CciElement*	newElement;
#ifdef _DEBUG
	CciSize	size;
	CciChar	name[256];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		name[0] = 0;
		size = cpiElementName(&rc, context->eParent, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		Log("\nPLUGIN: -> addItem(); %S, eParent %S (0x%x) bits(%d %d)\n",
            tagName, name, context->eParent, completePrev, completeNext);
	}
#endif

	// Create a new element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc,
					parser,
					CCI_ELEMENT_TYPE_NAME_VALUE,
					tagName,
					COMPLETE_BIT,
					COMPLETE_BIT);

	// now, we need to append the data, if any
	if (length > 0)
	{
		/* set the element value */
		cpiSetCharacterValueFromBuffer(&rc, newElement, dataIn, length);
	}

	// Add element to parent as last child
	cpiAddAsLastChild(&rc, context->eParent, newElement);

#ifdef _DEBUG
	if (traceModule) 
	{
		name[0] = 0;
		size = cpiElementName(&rc, context->eParent, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		Log("PLUGIN: <- addItem(); element %S (0x%x), eParent %S (0x%x) bits(%d %d)\n",
            tagName, newElement, name, context->eParent, completePrev, completeNext);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to parse the elements using a metadata file that has been loaded   */
/* into memory.  This routine will be entered recursively to handle children. */
/*                                                                            */
/******************************************************************************/

void parseElements(CciParser* parser, 
				   PARSER_CONTEXT_ST* context,
				   int	firstvar,
				   int	lastvar,
				   const CciChar * dataIn,
				   const CciChar * UnameArea,
				   VAR_FILE_REC *fvar)

{
	int	i;
	int j;
	int occOfs;
	int occur;
#ifdef _DEBUG
	int rc;
	CciSize size;
	CciChar name[512];
	CciBool	completePrev;
	CciBool completeNext;
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1); 
		Log("PLUGIN: -> parseElements(); element=0x%x %S bits(%d %d) firstvar=%d, lastvar=%d\n",
            context->eParent, name, completePrev, completeNext, firstvar, lastvar);
	}
#endif

	i=firstvar;
	while (i < lastvar)
	{
		// initialize offset of occurs items
		occOfs = 0;

		// get the number of occurences
		occur = 1;

		if (fvar[i].Occur > 1)
		{
			occur = fvar[i].Occur;
		}

		// initialize occurs counter
		j = 0;

		// check if this variable is an elementary item or a group item
		if (fvar[i].FirstChild == 0)
		{
			// check if we should skip this item
			if ((context->iFiller == 0) || ((fvar[i].Flags2 & FLAGS_FILLER) == 0))
			{
				// handle variables that occur more than once
				while (j < occur)
				{
					// build the line for an elementary item
					addItem(parser, 
							context, 
							UnameArea + fvar[i].Unameptr, 
							dataIn + fvar[i].Offset + occOfs,
							fvar[i].Length,
							fvar[i].Usage);

					// adjust the data area for the next occurence
					occOfs += fvar[i].Length;

					// go on to the next occurence
					j++;
				} // while (j < occur)
			}
			else
			{
#ifdef _DEBUG
				if (traceParser) 
				{
					Log("skipping filler element %S\n", UnameArea + fvar[i].Unameptr);
				}
#endif
			}
		}
		else
		{
			// handle variables that occur more than once
			while (j < occur)
			{
				// create the element as a name element
				context->eParent = createHeaderElement(parser, 
													   context, 
													   UnameArea + fvar[i].Unameptr);

				// Now, do all the children of the this element
				parseElements(parser, 
							  context, 
							  fvar[i].FirstChild, 
							  fvar[i].LastChild + 1,
							  dataIn + occOfs,
							  UnameArea,
							  fvar);

				// Now finish the name element
				finishHeaderElement(context);

				// adjust the data area for the next occurence
				occOfs += fvar[i].Length;

				// go on to the next occurence
				j++;
			} // while (j < occur)
		}

		// advance to the next element
		i = fvar[i].NextSibling;

		// check if we done
		if (i == 0)
		{
			// end the loop
			i = lastvar;
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		size = cpiElementName(&rc, context->eParent, name, sizeof(name) - 1); 
		Log("PLUGIN: <- parseElements(); element=0x%x %S bits(%d %d) firstvar=%d, lastvar=%d\n",
            context->eParent, name, completePrev, completeNext, firstvar, lastvar);
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to parse the next segment in the message.  The     */
/* corresponding definition file will be located and the      */
/* parseSegment routine will be called.                       */
/* segments to be parsed.                                     */
/*                                                            */
/**************************************************************/

void parseNextSegment(CciParser* parser,
					  PARSER_CONTEXT_ST* context)

{
	int	rc;
	int seglen=0;			// segment length in data
	int fileIdx=0;			// Index to file table
	char* vPtr;
	char segname[9];
	char charlen1[32];
	char charlen2[32];
	char charofs[32];
	char errmsg[512];
#ifdef _DEBUG
	CciChar	name[256];
	CciSize	length;
	CciBool	completePrev;
	CciBool completeNext;
#endif

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		memset(name, 0, sizeof(name));

		length = cpiElementName(&rc, context->eParent, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		Log("PLUGIN: -> parseNextSegment() context->eParent %S (0x%x) bits(%d %d)\n", 
			name, context->eParent, completePrev, completeNext);
	}
#endif

	// count the number of segments parsed
	segmentCount++;

	// clear the work areas
	memset(segname, 0, sizeof(segname));

	// get the segment length and name
	cciUcsToMbs(&rc, context->iBuffer + context->iOffset, segname, GROUPNAMELEN, DEF_CODEPAGE);
	seglen = getSegmentLength(context->iBuffer + context->iOffset, context->iOffset);

	// Tell what we are doing
#ifdef _DEBUG
	if (traceParser)
	{
		Log("\nProcessing message segment %s, length %d, offset %d\n", segname, seglen, context->iOffset);
	}
#endif

	// make sure that the segment length is reasonable
	if ((seglen == 0) || (seglen > (context->iLength - context->iOffset)))
	{
#ifdef _DEBUG
		if (debug)
		{
			Log("Invalid segment length %d, remaining %d\n", seglen, context->iLength - context->iOffset);
		}
#endif
	}

	// point to the 2TCG information for standard version and revision numbers
	vPtr = context->msgVersion;

	// check for message header or trailer groups
	if ((strcmp(segname, ACORDMHG) == 0) || (strcmp(segname, ACORDMTG) == 0))
	{
		if (strcmp(segname, ACORDMHG) == 0) 
		{
			// capture the version number for the header and trailer groups
			memcpy(context->hdrVersion, ACORDMHGSTD, ACORDSTDNOSIZE);
			cciUcsToMbs(&rc, 
						context->iBuffer + context->iOffset + ACORDMHGOFS, 
						context->hdrVersion + ACORDSTDNOSIZE, 
						ACORDVERNOSIZE,
						DEF_CODEPAGE);
		}

		if (strcmp(segname, ACORDMTG) == 0) 
		{
			if (context->iInTran == 1)
			{
				// mark the current transaction element as complete and
				// restore the body element as the parent of the next group
				finishHeaderElement(context);
				context->iInTran = 0;
			}
		}

		// point to the right standard number
		vPtr = context->hdrVersion;
	}

	// check for transaction header or control groups
	if ((strcmp(segname, ACORDTRG) == 0) || (strcmp(segname, ACORDTCG) == 0))
	{
		if (strcmp(segname, ACORDTRG) == 0)
		{
			// capture the version number for the header and trailer groups
			memcpy(context->trnVersion, ACORDMHGSTD, ACORDSTDNOSIZE);
			cciUcsToMbs(&rc, 
						context->iBuffer + context->iOffset + ACORDMHGOFS, 
						context->trnVersion + ACORDSTDNOSIZE, 
						ACORDVERNOSIZE,
						DEF_CODEPAGE);

			// check if we are in a transaction now
			if (context->iInTran == 1)
			{
				// mark the current transaction element as complete and
				// restore the body element as the parent of the next group
				finishHeaderElement(context);
			}

			// create a transaction element for this transaction
			context->eParent = createHeaderElement(parser, context, ACORDTRANNAMEL);

			// remember we are in a transaction
			context->iInTran = 1;
		}

		if (strcmp(segname, ACORDTCG) == 0)
		{
			cciUcsToMbs(&rc, context->iBuffer + context->iOffset + ACORDVERSIONOFFSET,
						context->msgVersion, sizeof(context->msgVersion), DEF_CODEPAGE);
		}

		// point to the right standard number
		vPtr = context->trnVersion;
	}

	// get the copy book file meta data, passing in the 
	// 4 character segment identifier in the message data
#ifndef _DEBUG
	fileIdx = getMetaData(0, vPtr, segname, context->iBuffer[context->iOffset + GROUPVERSIONOFS]);
#else
	fileIdx = getMetaData(traceParser, vPtr, segname, context->iBuffer[context->iOffset + GROUPVERSIONOFS]);
#endif


	if (fileIdx < 0)
	{
		// didn't find the file meta data
		// write an error message
		sprintf(errmsg, "Error getting meta data for segment %s at offset %d\n", segname, context->iOffset); 

#ifdef _DEBUG
		if (debug)
		{
			Log(errmsg);
		}
#endif

		charofs[0] = 0;
		itoa(context->iOffset, charofs, 10);
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "parseMessage",
						  ERRDICT,
						  ACORD_META_NOT_FOUND,
						  errmsg,
						  segname,
						  charofs,
						  0);
	}
	else
	{
#ifdef _DEBUG
		if (traceParser)
		{
			Log("First variable name is %S, length %d\n",
				fileTable[fileIdx].UnamePtr + fileTable[fileIdx].fvar[0].Unameptr,
				fileTable[fileIdx].fvar[0].Length);
		}
#endif

		// check if the lengths match
		if ((fileTable[fileIdx].fvar[0].Length) != seglen)
		{
			// build error message
			sprintf(errmsg, "Segment length (%d) does not match copybook length (%d)\n", 
					seglen, fileTable[fileIdx].fvar[0].Length);
#ifdef _DEBUG
			if (debug)
			{
				Log(errmsg);
			}
#endif

			charlen1[0] = 0;
			itoa(seglen, charlen1, 10);
			charlen2[0] = 0;
			itoa(fileTable[fileIdx].fvar[0].Length, charlen2, 10);
			charofs[0] = 0;
			itoa(context->iOffset, charofs, 10);

			cciThrowException(&rc, 
							  CCI_PARSER_EXCEPTION, 
							  __FILE__, 
							  __LINE__, 
							  "parseMessage",
							  ERRDICT,
							  ACORD_SEGMENT_LENGTH_BAD,
							  &errmsg[0],
							  &charlen1, 
							  &charlen2,
							  &segname,
							  &charofs,
							  0);
		}


		// Process the data and the definitions
		parseElements(parser, 
					  context,
					  0,
					  fileTable[fileIdx].maxvar,
					  context->iBuffer + context->iOffset,
					  fileTable[fileIdx].UnamePtr,
					  fileTable[fileIdx].fvar);

			// Decrement the file usage counter
#ifdef _DEBUG
		decrementUseCount(traceParser, fileIdx);
#else
		decrementUseCount(0, fileIdx);
#endif
	}

#ifdef _DEBUG
	if (traceParser)
	{
		Log("Segment %s parsed length %d offset %d\n", 
			segname, seglen, context->iOffset);
	}
#endif

	// Check if this is the first time through
	if (context->iOffset == 0)
	{
		// mark the body root first child flag as complete
		cpiSetElementCompletePrevious(&rc, context->eBody, COMPLETE_BIT);

#ifdef _DEBUG
		if (traceParser)
		{
			Log("Body element first child flag set to complete rc=%d body=0x%x\n", 
				rc, context->eBody);
		}
#endif
	}

	// move on to the next segment in the input buffer
	context->iOffset += seglen;

	// skip any new line or carriage return characters found between segments
	while ((context->iOffset < context->iLength) && 
		(context->iBuffer[context->iOffset] < chSpace))
	{
		context->iOffset++;
	} // while

	if (context->iOffset == context->iLength)
	{
		// mark the body root last child as complete
		cpiSetElementCompleteNext(&rc, context->eBody, COMPLETE_BIT);

		// mark the last child of the body as complete as well
		cpiSetElementCompleteNext(&rc, context->eParent, COMPLETE_BIT);
	}

#ifdef _DEBUG
	if (traceParser)
	{
		// dump out all the elements at this point
		dumpAllElements(context->eBody, 0);

		memset(name, 0, sizeof(name));

		length = cpiElementName(&rc, context->eParent, name, sizeof(name));
		completePrev = cpiElementCompletePrevious(&rc, context->eParent);
		completeNext = cpiElementCompleteNext(&rc, context->eParent);
		Log(" context->eParent %S (0x%x) bits(%d %d)\n", 
			name, context->eParent, completePrev, completeNext);
	}
#endif

#ifdef _DEBUG
	if (traceModule)
	{
	    Log("PLUGIN: <- parseNextSegment() offset = %d, length = %d\n",
		        context->iOffset, context->iLength);
	}
#endif
}

const char *  getAcordsubLevel()

{
	return Level;
}

//////////////////////
// End of file
//////////////////////