//  mgmtmsg.cpp 

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

#ifdef WIN32
#include <windows.h>
#include "largeint.h"
#include <wchar.h>
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCpi.h>
#include <BipCni.h>

#include "BipSampPluginUtil.h"
#include "parsubs.h"
#include "trace.h"
#include "mtdstruc.h"
#include "metamem.h"
#include "metadata.h"
#include "miscsubs.h"
#include "SegParse.h"
#include "acordsub.h"
#include "statfile.h"
#include "context.h"
#include "common.h"

// Parser exception codes (built by message compiler)
#include "acorderr.h"

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#ifdef WIN32
#define MESSAGE_SET		L"MessageSet"
#define MGMT_MSG		L"SYSTMGMT"
#endif

#define FLUSH			"FLUSH   "
#define FLUSHALL		"FLUSHALL"
#define CACHSTAT		"CACHSTAT"
#define STATS			"STATS   "
#define TRACEON			"TRACEON "
#define TRACEOFF		"TRACEOFF"
#define TRACSTAT		"TRACSTAT"
#define DUMPSTAT		"DUMPSTAT"
#define GETLEVEL		"GETLEVEL"
#define TYPE_LENGTH		8
#define FLUSH_SIZE		8
#define MIN_MSG_LEN		16
#define ACORDSTATSFILE	"ACORD_STAT_FILE"

#ifdef _DEBUG
static char Level[]="Mgmtmsg.cpp V1.0 Debug version("__DATE__" "__TIME__")";
#else
static char Level[]="Mgmtmsg.cpp V1.0 Release version("__DATE__" "__TIME__")";
#endif

#ifdef _DEBUG
	extern int debug;
	extern int traceParser;
	extern int traceWrite;
	extern int traceModule;
	extern int traceMgmt;
#endif

	extern FILETBL* fileTable;		// File table to cache recent files 

	extern int	parserStartHigh;	// Time parser was started (high word)
	extern int	parserStartLow;		// Time parser was started (low word)
	extern int	totalParseTime;		// Total parsing time in milliseconds
	extern int	totalWriteTime;		// Total time to build a buffer in milliseconds
	extern int	maxParseTime;		// Maximum time to parse a message
	extern int	minParseTime;		// Minimum time to parse a message
	extern int	maxWriteTime;		// Maximum time to build a buffer
	extern int	minWriteTime;		// Minimum time to build a buffer
	extern int	totalBytes;			// Number of bytes in all messages processed
	extern int	totalUncomp;		// Number of bytes in all input after decompression
	extern int	maxMessage;			// Largest input message processed
	extern int	minMessage;			// Smallest input message processed
	extern int	totalWritten;		// Number of bytes written
	extern int	maxWriteLength;		// Largest message written
	extern int	minWriteLength;		// Smallest message written

	extern int	messageCount;		// Number of messages processed    
	extern int	segmentCount;		// Number of group segments processed
	extern int	writeCount;			// Number of cpiWriteBuffer calls processed
	extern int	compressMsgs;		// Number of compressed messages processed    
	extern int	badMessageCount;	// Number of invalid messages rejected

	extern int	useCounter;			// Counter for least recently used meta data files
	extern int	filesOpened;		// files added to cache
	extern int	filesRemoved;		// files removed from cache
	extern int	filesMax;			// Maximum number of files in cache

	extern int	packedTruncated;	// Packed fields truncated
	extern int	displayTruncated;	// Display fields truncated
	extern int	numericTruncated;	// Numeric fields truncated
	extern int	binaryTruncated;	// Binary fields truncated


/******************************************************************************/
/*                                                                            */
/* Routine to process a cache flush management message.                    .  */
/*                                                                            */
/******************************************************************************/

void processFlushMsg(int ver, int mod, char * fileCodes, int length)

{
	int	i;
	int	j;
	int	count;
	char segName[FLUSH_SIZE + 1];

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
	    Log("PLUGIN: -> processFlushMsg()\n");
	}
#endif

	if (length >= FLUSH_SIZE)
	{
		// calculate the number of entries in the message
		count = length / FLUSH_SIZE;

		i = 0;
		while (i < count)
		{
			// get the segment name as a string
			j = 0;
			while ((j < FLUSH_SIZE) && (fileCodes[j] > ' '))
			{
				segName[j] = fileCodes[j];
				j++;
			}

			// terminate the string
			segName[j] = 0;

			removeMetaDataEntry(segName);

			// move on to the next entry
			fileCodes += FLUSH_SIZE;
			i++;
		}
	}
	else
	{
		// Invalid message, no entries to flush
#ifdef _DEBUG
		if (traceModule)
		{
			Log("No entries - length(%d)\n", length);
		}
#endif
	}

	// Finally, flush the standards and defaults files
	initializeStandards();

#ifdef _DEBUG
	if (traceModule)
	{
	    Log("PLUGIN: <- processFlushMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a cache flush all management message.                .  */
/*                                                                            */
/******************************************************************************/

void processFlushAllMsg()

{
	// Flush all the metadata entries
	removeAllMetaData();
}

/******************************************************************************/
/*                                                                            */
/* Routine to add a name element to the parse tree.                        .  */
/*                                                                            */
/******************************************************************************/

CciElement * newNameElement(CciParser* parser, const char* name)

{
	int			rc;
	CciElement* newElement;
	CciChar		Uname[64];
	CciSize		size;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> newNameElement - creating name element for (%s)\n", name);
	}
#endif

	// translate the input name to Unicode
	size = cciMbsToUcs(&rc, name, Uname, sizeof(Uname) - 1, DEF_CODEPAGE);

	// Create a new name element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc, 
											   parser, 
											   CCI_ELEMENT_TYPE_NAME, 
											   Uname, 
											   COMPLETE_BIT,
											   COMPLETE_BIT);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- newNameElement - created name element 0x%x\n", newElement);
	}
#endif

	return newElement;
}

///////////////////////////////////////////////////
//
// Routine to add an element with a value 
// given as a string, and with
// the given name, under the
// specified parent element.
//
///////////////////////////////////////////////////

void addCharacterElement(CciParser* parser, 
						 const char * elementValue,
						 const char * elementName,
						 CciElement *parent) 

{
	int			rc;
	int			size;
	CciChar		workValue[1024];
	CciChar		Uname[64];
	CciElement*	newElement;

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> addCharacterElement - element %s as child of 0x%x\n", elementName, parent);
	}
#endif

	// translate the input name to Unicode
	size = cciMbsToUcs(&rc, elementName, Uname, sizeof(Uname) - 1, DEF_CODEPAGE);

	// Create a new name element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc, 
											   parser, 
											   CCI_ELEMENT_TYPE_NAME_VALUE,
											   Uname,
											   COMPLETE_BIT,
											   COMPLETE_BIT);

	// Copy the text into a wide ImbString 
	workValue[0] = 0;
	size = cciMbsToUcs(&rc, elementValue, workValue, sizeof(workValue) - 1, DEF_CODEPAGE);

	// Set the element value
	cpiSetElementCharacterValue(&rc, newElement, workValue, size);

	// add the element to the tree
	cpiAddAsLastChild(&rc, parent, newElement);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- addCharacterElement - element %s created with value (%s)\n", 
			elementName, elementValue);
	}
#endif
}

///////////////////////////////////////////////////
//
// Routine to add an element with a value 
// given as an integer, and with
// the given name, under the
// specified parent element.
//
///////////////////////////////////////////////////

void addStatElement(CciParser* parser, 
					const int value, 
					const char* name, 
					CciElement* parent)

{
	int			rc;
	CciSize		size;
	CciElement* newElement;
	CciChar		UvalueStr[256];
	CciChar		Uname[64];
	char		valueChar[32];

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> addStatElement - element for (%s)\n", name);
	}
#endif

	// translate the input name to Unicode
	size = cciMbsToUcs(&rc, name, Uname, sizeof(Uname) - 1, DEF_CODEPAGE);

	// Create a new element for this tag 
	// Create a new name element for the tag 
	newElement = cpiCreateAndInitializeElement(&rc, 
											   parser, 
											   CCI_ELEMENT_TYPE_NAME_VALUE, 
											   Uname, 
											   COMPLETE_BIT,
											   COMPLETE_BIT);

	// now, add the value element
	itoa(value, valueChar, 10);

	// Copy the text into a wide ImbString 
	UvalueStr[0] = 0;
	size = cciMbsToUcs(&rc, valueChar, UvalueStr, sizeof(UvalueStr) - 1, DEF_CODEPAGE);

	// Set the element value
	cpiSetElementCharacterValue(&rc, newElement, UvalueStr, size);

	// add the element to the tree
	cpiAddAsLastChild(&rc, parent, newElement);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- addStatElement - element 0x%x created as child of 0x%x for (%S)\n", 
			newElement, parent, Uname);
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a statistics management message.                     .  */
/*                                                                            */
/*  Build a parse tree of the following format (as if in XML)                 */
/*   <ACORD>                                                                  */
/*    <statistics>                                                            */
/*     <cache>                                                                */
/*      <filesOpened>nnn</filesOpened>                                        */
/*      <filesRemoved>nnn</filesRemoved>                                      */
/*      <filesMax>nnn</filesMax>                                              */
/*      <useCounter>nnn</useCounter>                                          */
/*     </cache>                                                               */
/*     <messages>                                                             */
/*      <messageCount>nnn</messageCount>                                      */
/*      <compressedMessages>nnn</compressedMessages>                          */
/*      <segmentCount>nnn</segmentCount>                                      */
/*      <writeCount>nnn</writeCount>                                          */
/*      <invalidMsgCount>nnn</invalidMsgCount>                                */
/*      <maxMessageSize>nnn</maxMessageSize>                                  */
/*      <minMessageSize>nnn</minMessageSize>                                  */
/*      <writeBufferCount>nnn</writeBufferCount>                              */
/*      <maxOutputMsg>nnn</maxOutputMsg>                                      */
/*      <minOutputMsg>nnn</minOutputMsg>                                      */
/*      <averageOutputMsg>nnn</averageOutputMsg>                              */
/*      <averageMessageSize>nnn</averageMessageSize>                          */
/*      <averageUncompressedSize>nnn</averageUncompressedSize>                */
/*      <displayFieldsTruncated>nnn</displayFieldsTruncated>                  */
/*      <numericFieldsTruncated>nnn</numericFieldsTruncated>                  */
/*      <averageParseTime>nnn</averageParseTime>                              */
/*      <maxParseTime>nnn</maxParseTime>                                      */
/*      <minParseTime>nnn</minParseTime>                                      */
/*      <averageWriteTime>nnn</averageWriteTime>                              */
/*      <maxWriteTime>nnn</maxWriteTime>                                      */
/*      <minWriteTime>nnn</minWriteTime>                                      */
/*     </messages>                                                            */
/*    </statistics>                                                           */
/*   </ACORD>                                                                 */
/*                                                                            */
/******************************************************************************/

void processStatsMsg(CciParser* parser, int ver, int mod)

{
	int			rc;
	int			average;
	int			averageTime=0;
	CciElement* body;
	CciElement* nameElement;
	CciElement* statisticsElement;
	CciElement* cacheElement;
	CciElement* messagesElement;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> processStatsMsg() ver %d mod %d\n", ver, mod);
	}
#endif

	// Get a pointer to the root element for the parser
	body = cpiRootElement(&rc, parser);

	// create the statistics element
	nameElement = newNameElement(parser, "Statistics");
	cpiAddAsFirstChild(&rc, body, nameElement);
	statisticsElement = nameElement;

	// create the cache element
	nameElement = newNameElement(parser, "cache");
	cpiAddAsFirstChild(&rc, statisticsElement, nameElement);
	cacheElement = nameElement;

	// create the messages element under statistics
	messagesElement = newNameElement(parser, "messages");
	cpiAddAfter(&rc, cacheElement, messagesElement);

	addStatElement(parser, filesOpened, "filesOpened", cacheElement);
	addStatElement(parser, filesRemoved, "filesRemoved", cacheElement);
	addStatElement(parser, filesMax, "filesMax", cacheElement);
	addStatElement(parser, useCounter, "useCounter", cacheElement);

	addStatElement(parser, messageCount, "messageCount", messagesElement);
	addStatElement(parser, compressMsgs, "compressedMessages", messagesElement);
	addStatElement(parser, segmentCount, "segmentCount", messagesElement);
	addStatElement(parser, writeCount, "writeCount", messagesElement);
	addStatElement(parser, badMessageCount, "invalidMsgCount", messagesElement);

	addStatElement(parser, maxMessage, "maxMessageSize", messagesElement);
	addStatElement(parser, minMessage, "minMessageSize", messagesElement);

	average = 0;
	if (messageCount > 0)
	{
		average = totalBytes / messageCount;
	}

	addStatElement(parser, average, "averageMessageSize", messagesElement);

	average = 0;
	if (messageCount > 0)
	{
		average = totalUncomp / messageCount;
	}

	addStatElement(parser, average, "averageUncompressedSize", messagesElement);

	addStatElement(parser, writeCount, "writeBufferCount", messagesElement);
	addStatElement(parser, maxWriteLength, "maxOutputMsg", messagesElement);
	addStatElement(parser, minWriteLength, "minOutputMsg", messagesElement);

	average = 0;
	if (writeCount > 0)
	{
		average = totalWritten / writeCount;
	}

	addStatElement(parser, average, "averageOutputMsg", messagesElement);

	addStatElement(parser, displayTruncated, "displayFieldsTruncated", messagesElement);
	addStatElement(parser, numericTruncated, "numericFieldsTruncated", messagesElement);

	// get the count as a string
	averageTime = 0;
	if (messageCount > 0)
	{
		averageTime = totalParseTime / messageCount;
	}

	addStatElement(parser, averageTime, "averageParseTime", messagesElement);
	addStatElement(parser, maxParseTime, "maxParseTime", messagesElement);
	addStatElement(parser, minParseTime, "minParseTime", messagesElement);

	// now, add the value element
	averageTime = 0;
	if (messageCount > 0)
	{
		averageTime = totalWriteTime / messageCount;
	}

	addStatElement(parser, averageTime, "averageWriteTime", messagesElement);
	addStatElement(parser, maxWriteTime, "maxWriteTime", messagesElement);
	addStatElement(parser, minWriteTime, "minWriteTime", messagesElement);

	// Make sure the top level is set to complete
	cpiSetElementCompletePrevious(&rc, body, COMPLETE_BIT);
	cpiSetElementCompleteNext(&rc, body, COMPLETE_BIT);

#ifdef _DEBUG
	if (traceModule)
	{
		dumpAllElements(body, 1);
		Log("PLUGIN: <- processStatsMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Write the statistics out to a file.                                        */
/*                                                                            */
/* This function requires an environment variable (ACORD_STATS_FILE) to be    */
/* set to the full path and file name of a statistics file.  A statistics     */
/* record will be appended to the file pointed to by this variable.           */
/*                                                                            */
/******************************************************************************/

void dumpStatistics()

{
	char*		fileVar;
	FILE*		statfile;
	STAT_REC	stat;
  	SYSTEMTIME		curtime;
	FILETIME		filtime;

	// Get the current time
	GetSystemTime(&curtime);
	SystemTimeToFileTime(&curtime, &filtime);

	fileVar = getenv(ACORDSTATSFILE);
	if (fileVar != NULL)
	{
		// try and open the statistics file
		statfile = fopen(fileVar, "w+");
		if (statfile != NULL)
		{
			/* clear the statistics record */
			memset(&stat, 0, sizeof(stat));

			/* build the statistics file record */
			stat.version = 0;
			stat.startTimeHigh = parserStartHigh;
			stat.startTimeLow = parserStartLow;
			stat.endTimeHigh = filtime.dwHighDateTime;
			stat.endTimeLow  = filtime.dwLowDateTime;
			stat.messageCount = messageCount;
			stat.segmentCount = segmentCount;
			stat.writeCount = writeCount;
			stat.compressMsgs = compressMsgs;
			stat.badMessageCount = badMessageCount;
			stat.totalBytes = totalBytes;
			stat.maxMessage = maxMessage;
			stat.minMessage = minMessage;
			stat.totalWritten = totalWritten;
			stat.maxWriteLength = maxWriteLength;
			stat.minWriteLength = minWriteLength;
			stat.totalParseTime = totalParseTime;
			stat.totalWriteTime = totalWriteTime;
			stat.maxParseTime = maxParseTime;
			stat.maxParseTime = minParseTime;
			stat.maxWriteTime = maxWriteTime;
			stat.minWriteTime = minWriteTime;
			stat.useCounter = useCounter;
			stat.filesOpened = filesOpened;
			stat.filesRemoved = filesRemoved;
			stat.filesMax = filesMax;
			stat.displayTruncated = displayTruncated;
			stat.numericTruncated = numericTruncated;

			/* Now, write out the statistics */
			fwrite(&stat, 1, sizeof(stat), statfile);

			// close the file
			fclose(statfile);
		}
	}
}

/******************************************************************************/
/*                                                                            */
/* Routine to build a parse tree with the current cache status.            .  */
/*                                                                            */
/******************************************************************************/

void buildCacheMsg(CciParser* parser)

{
	int			rc;
	int			i;
	CciElement* body;
	CciElement* cacheElement;
	CciElement* nameElement;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> buildCacheMsg()\n");
	}
#endif

	// Get a pointer to the body element for the parser
	body = cpiRootElement(&rc, parser);

	// create the cache status element
	cacheElement = newNameElement(parser, "CacheStatus");
	cpiAddAsFirstChild(&rc, body, cacheElement);

	i = 0;
	while (i < MAX_FILES)
	{
		// is this a valid entry?
		if ((fileTable[i].fileName[0] > 0) && (fileTable[i].fileName[0] != DELETED_FILE_BYTE))
		{
			// create an entry for this metadata cache file
			nameElement = newNameElement(parser, "CacheEntry");
			cpiAddAsLastChild(&rc, cacheElement, nameElement);

			// add the children to this entry
			addStatElement(parser, i, "Index", nameElement);
			addCharacterElement(parser, fileTable[i].fileName, "FileName", nameElement);
			addStatElement(parser, fileTable[i].dataLength , "Length", nameElement);
			addStatElement(parser, fileTable[i].inUseCount, "UseCount", nameElement);
			addStatElement(parser, fileTable[i].flags, "Flags", nameElement);
		}

		i++;
	}

	// Make sure the body element is set to complete
	cpiSetElementCompletePrevious(&rc, body, COMPLETE_BIT);
	cpiSetElementCompleteNext(&rc, body, COMPLETE_BIT);

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- buildCacheMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to build a parse tree with the current trace status.            .  */
/*                                                                            */
/******************************************************************************/

void buildTraceStatusMsg(CciParser* parser)

{
	int			rc;
	CciElement* body;
	CciElement* traceElement;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> buildTraceStatusMsg()\n");
	}
#endif

	// Get a pointer to the body element for the parser
	body = cpiRootElement(&rc, parser);

	// create the trace status element
	traceElement = newNameElement(parser, "TraceStatus");
	cpiAddAsFirstChild(&rc, body, traceElement);

#ifdef _DEBUG
	// add the individual trace status entries to the parse tree
	addStatElement(parser, debug, "debug", traceElement);
	addStatElement(parser, traceModule, "traceModule", traceElement);
	addStatElement(parser, traceParser, "traceParser", traceElement);
	addStatElement(parser, traceWrite, "traceWrite", traceElement);
	addStatElement(parser, traceMgmt, "traceMgmt", traceElement);
#else
	addCharacterElement(parser, "No trace available", "traceType", traceElement);
#endif

	// Make sure the body element is set to complete
	cpiSetElementCompletePrevious(&rc, body, COMPLETE_BIT);
	cpiSetElementCompleteNext(&rc, body, COMPLETE_BIT);


	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- buildTraceStatusMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a request to turn tracing on.                        .  */
/*                                                                            */
/******************************************************************************/

void processTraceOnMsg(int ver, int mod, int length, char * traceTypes)

{
#ifdef _DEBUG
	int	count=0;

	// Tell what we are doing
	if (traceModule)
	{
		Log("PLUGIN: -> processTraceOnMsg() ver %d mod %d\n", ver, mod);
	}

	/* calculate the number of requests in the message */
	if (length > 0)
	{
		count = length / 8;
	}

	/* Turn on the requested traces */
	processTraceMessage(traceTypes, count, 1);

	if (traceModule)
	{
		Log("PLUGIN: <- processTraceOnMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a request to turn tracing off.                       .  */
/*                                                                            */
/******************************************************************************/

void processTraceOffMsg()

{
#ifdef _DEBUG
	char tempMsg[] = TRACEOFF;

	// Tell what we are doing
	if (traceModule)
	{
		Log("PLUGIN: -> processTraceOffMsg()\n");
		Log("Closing trace file\n");
	}

	processTraceMessage(tempMsg, 1, 0);

	if (traceModule)
	{
		Log("PLUGIN: <- processTraceOffMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process a request to get the module levels.                  .  */
/*                                                                            */
/*  Build a parse tree of the following format (as if in XML)                 */
/*   <ACORD>                                                                  */
/*    <ModuleLevels>                                                          */
/*     <Acordsub>nnn</Acordsub>                                               */
/*     <Metadata>nnn</Metadata>                                               */
/*     <Mgmtmsg>nnn</Mgmtmsg>                                                 */
/*     <Miscsubs>nnn</Miscsubs>                                               */
/*     <Parsubs>nnn</Parsubs>                                                 */
/*     <SegParse>nnn</SegParse>                                               */
/*     <Trace>nnn</Trace>                                                     */
/*    </ModuleLevels>                                                         */
/*   </ACORD>                                                                 */
/******************************************************************************/

void processGetlevelMsg(CciParser* parser)

{
	int			rc;
	const char*	moduleLevel;
	CciElement* body;
	CciElement* levelElement;

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> processGetlevelMsg()\n");
	}
#endif

	// Get a pointer to the body element for the parser
	body = cpiRootElement(&rc, parser);

	// create the module level element
	// Create a new name element for the tag 
	levelElement = newNameElement(parser, "ModuleLevels");
	cpiAddAsFirstChild(&rc, body, levelElement);

	moduleLevel = getAcordsubLevel();
	addCharacterElement(parser, moduleLevel, "Acordsub", levelElement);

	moduleLevel = getMetadataLevel();
	addCharacterElement(parser, moduleLevel, "Metadata", levelElement);

	addCharacterElement(parser, Level, "mgmtmsg", levelElement);

	moduleLevel = getMiscsubsLevel();
	addCharacterElement(parser, moduleLevel, "Miscsubs", levelElement);

	moduleLevel = getParsubsLevel();
	addCharacterElement(parser, moduleLevel, "Parsubs", levelElement);

	moduleLevel = getSegParseLevel();
	addCharacterElement(parser, moduleLevel, "Segparse", levelElement);

	moduleLevel = getTraceLevel();
	addCharacterElement(parser, moduleLevel, "Trace", levelElement);

	// Make sure the top level is set to complete
	cpiSetElementCompletePrevious(&rc, body, COMPLETE_BIT);
	cpiSetElementCompleteNext(&rc, body, COMPLETE_BIT);

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		dumpAllElements(body, 1);
		Log("PLUGIN: <- processGetlevelMsg()\n");
	}
#endif
}

/******************************************************************************/
/*                                                                            */
/* Routine to process messages that should be processed immediately.       .  */
/*                                                                            */
/******************************************************************************/

void processMgmtMsg(CciParser* parser, PARSER_CONTEXT_ST* context) 

{
	int		rc;
	char*	buffer;
	char	msgType[TYPE_LENGTH + 1];
	char	version[5];
	char	modLevel[5];
	char	errmsg[512];
	char	charlen[32];
 
	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> processMgmtMsg() length %d\n", context->iDataLen);
	}
#endif

	if (context->iDataLen < MIN_MSG_LEN)
	{
		// prepare message areas
		charlen[0] = 0;
		itoa(context->iDataLen, charlen, 10);
		sprintf(errmsg, "Management message too short (%d)", context->iDataLen);

		// throw an exception - message is too short
		cciThrowException(&rc, 
						  CCI_PARSER_EXCEPTION, 
						  __FILE__, 
						  __LINE__, 
						  "getSegmentLength",
						  ERRDICT,
						  ACORD_MGMT_TOO_SHORT,
						  errmsg,
						  charlen,
						  0);
	}

	// allocate an area of memory for a copy of the message in ASCII
	buffer = (char *)malloc(context->iDataLen + 1);
	buffer[context->iDataLen] = 0;

	// translate the message to ASCII from Unicode
	cciUcsToMbs(&rc, context->iBuffer, buffer, context->iDataLen, DEF_CODEPAGE);

	// get the type of management message
	memcpy(msgType, buffer + 8, TYPE_LENGTH);
	msgType[TYPE_LENGTH] = 0;

	// get the version number and modification level
	memcpy(version, buffer, 4);
	version[4] = 0;
	context->iMgmtVersion = atoi(version);
	memcpy(modLevel, buffer + 4, 4);
	modLevel[4] = 0;
	context->iMgmtModLevel = atoi(modLevel);

	if ((strcmp(version, "0000") == 0) && (strcmp(modLevel, "0000") == 0)) 
	{
		if (strcmp(msgType, FLUSH) == 0)
		{ 
			processFlushMsg(context->iMgmtVersion, 
							context->iMgmtModLevel, 
							buffer + 16, 
							context->iDataLen - 16);
		}

		if (strcmp(msgType, FLUSHALL) == 0)
		{
			processFlushAllMsg();
		}

		if (strcmp(msgType, DUMPSTAT) == 0)
		{
			// write the statistics to a file
			dumpStatistics();
		}

		if (strcmp(msgType, TRACEON) == 0)
		{
			processTraceOnMsg(context->iMgmtVersion, 
							  context->iMgmtModLevel, 
							  context->iDataLen - 16, 
							  buffer + 16);
		}

		if (strcmp(msgType, TRACEOFF) == 0)
		{
			processTraceOffMsg();
		}
	}
	else
	{
		// wrong version or modification level
#ifdef _DEBUG
		if (traceMgmt)
		{
			Log("Wrong version(%s) or modification level(%s)\n", version, modLevel);
		}
#endif
	}

	// release the ASCII buffer
	free(buffer);

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- processMgmtMsg()\n");
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Routine to build a parse tree appropriate for the type of message.      .  */
/*                                                                            */
/******************************************************************************/

void parseMgmtMsg(CciParser* parser, PARSER_CONTEXT_ST* context)

{
	int		rc;
	CciElement*	body;
	char	msgType[TYPE_LENGTH + 1];

	// Tell what we are doing
#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> parseMgmtMsg() length %d\n", context->iDataLen);
	}
#endif

	// translate the message type to ASCII from Unicode
	cciUcsToMbs(&rc, context->iBuffer + 8, msgType, TYPE_LENGTH, DEF_CODEPAGE);
	msgType[TYPE_LENGTH] = 0;

		// Tell what we are doing
#ifdef _DEBUG
	if (traceMgmt)
	{
		Log("parsing management message type(%s) version(%d) modification level(%d)\n", 
			msgType, context->iMgmtVersion, context->iMgmtModLevel);
	}
#endif

	if ((strcmp(msgType, STATS) == 0) || (strcmp(msgType, DUMPSTAT) == 0))
	{
		// build a statistics message
		processStatsMsg(parser, context->iMgmtVersion, context->iMgmtModLevel);
	}

	if (strcmp(msgType, GETLEVEL) == 0)
	{
		// build a parser level message
		processGetlevelMsg(parser);
	}

	if ((strcmp(msgType, FLUSH) == 0) || (strcmp(msgType, FLUSHALL) == 0) ||
		(strcmp(msgType, CACHSTAT) == 0))
	{
		// build a message with the current contents of the metadata cache
		buildCacheMsg(parser);
	}

	if ((strcmp(msgType, TRACEON) == 0) || (strcmp(msgType, TRACEOFF) == 0) ||
		(strcmp(msgType, TRACSTAT) == 0))
	{
		// build a message with the current trace status
		buildTraceStatusMsg(parser);
	}

	// finally, we must mark the body element complete, in case
	// some node tries to use this message
	body = cpiRootElement(&rc, parser);
	if (body != CCI_NULL_ADDR)
	{
		cpiSetElementCompleteNext(&rc, body, COMPLETE_BIT);
		cpiSetElementCompletePrevious(&rc, body, COMPLETE_BIT);
	}

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- parseMgmtMsg()\n");
	}
#endif

	return;
}

/******************************************************************************/
/*                                                                            */
/* Check for system management message by checking the message set and type.  */
/*                                                                            */
/******************************************************************************/

int checkForMgmtMsg(CciParser* parser)

{
	int			rc=0;
	int			RC;
	int			size;
	CciElement*	root;
	CciElement*	body;
	CciElement*	MSet;
	CciChar		UMessageSet[128];

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: -> checkForMgmtMsg\n");
	}
#endif

	/* Initialize work areas */
	UMessageSet[0] = 0;
	//memset(UMessageSet, 0, sizeof(UMessageSet));

	/* find the body element for this parser */
	body = cpiRootElement(&RC, parser);

	/* check if we found the parser body */
	if (RC == CCI_SUCCESS)
	{
		// Now, find the real root element
		root = cpiParent(&RC, body);

		/* check if we found the root element */
		if (RC == CCI_SUCCESS)
		{
			/* find the properties as the first child of the root (Properties) */
			MSet = cniFirstChild(&RC, root);

			if (RC == CCI_SUCCESS)
			{
				// find the first property, using the node utility function
				// to force the properties to be parsed
				MSet = cniFirstChild(&RC, MSet);
			}

			/* check if we found the properties */
			if (RC == CCI_SUCCESS)
			{
				MSet = findNextElement(MSet, MESSAGE_SET);

				if (MSet != CCI_NULL_ADDR)
				{
					// found the property, get the value
					size = cpiElementCharacterValue(&RC, MSet, UMessageSet, sizeof(UMessageSet));

					// check if this is a management message
					if ((RC == CCI_SUCCESS) && (CciCharCmp(UMessageSet, MGMT_MSG) == 0))
					{
						// This is a management message, now we must set the return code 
						rc = 1;
					}
				}
				else
				{
					// couldn't find MessageSet element
#ifdef _DEBUG
					if (traceMgmt)
					{
						Log("Unable to locate MessageSet element\n");
					}
#endif
				}
			}
			else
			{	
				// couldn't find properties element
#ifdef _DEBUG
				if (traceMgmt)
				{
					Log("Unable to locate properties element RC %d \n", RC);
				}
#endif
			}
		}
		else
		{	
			// couldn't find parser root element
#ifdef _DEBUG
			if (traceMgmt)
			{
				Log("Unable to locate root element RC %d \n", RC);
			}
#endif
		}
	}
	else
	{
		// couldn't find parser root element
#ifdef _DEBUG
		if (traceMgmt)
		{
			Log("Unable to locate parser root element RC %d \n", RC);
		}
#endif
	}

#ifdef _DEBUG
	if (traceModule)
	{
		Log("PLUGIN: <- checkForMgmtMsg rc %d \n", rc);
	}
#endif

	return rc;
}

//////////////////////
// End of file
//////////////////////